local _savedAccountVariables
local _savedLocalVariables
local _availablePersonalities = {}
local _availableEmotes = {}
local _numberOfAvailableEmotes = 0
local _idEmoteList = {}
local _emoteData
local _categoryNames = {}
local _mainWindow = {}
local _personalityList = {}
local _emoteSelectionWindow = {}
local _isEmoteSynchronizationActive = false
local _name = "LovelyEmotes"

LovelyEmotes = {
	Name = _name,
	Version = "4.27",

	Modifications = {},
	AvailableEmotes = _availableEmotes,
	AvailablePersonalities = _availablePersonalities,
	MainWindow = _mainWindow,
	RadialMenu = {},
	PersonalityList = _personalityList,
	EmoteSelectionWindow = _emoteSelectionWindow,
}

function LovelyEmotes.GetNumberOfAvailableEmotes()
	return _numberOfAvailableEmotes
end

local function OnGameCameraUiModeChanged(eventCode)
	_emoteSelectionWindow.HideWindow()

	if _savedAccountVariables.MinimizeAutomatically and not IsGameCameraUIModeActive() then
		if _mainWindow.IsMinimized() then
			_mainWindow.SetMinimizedLocked(true)
			return
		end

		_mainWindow.SetMinimized(true, true)
	end
end

local function OnCombatStateChanged(eventCode, inCombat)
	if _mainWindow.IsMinimizedLocked() and not inCombat then return end
	_mainWindow.SetMinimized(inCombat, false)
end

local function SortAvailableEmotes()
	if _savedAccountVariables.ShowSlashNames then
		table.sort(_availableEmotes, function(firstValue, secondValue) return firstValue.SlashName < secondValue.SlashName end)
	else
		table.sort(_availableEmotes, function(firstValue, secondValue) return firstValue.DisplayName < secondValue.DisplayName end)
	end
end
LovelyEmotes.SortAvailableEmotes = SortAvailableEmotes

local function RefreshAvailableEmotesOverridden()
	for id, emote in pairs(_idEmoteList) do
		if id > -1 then emote.IsOverridden = IsPlayerEmoteOverridden(id) end
	end
end

local function ConcurrencyWorkaround()
	if CustomIdleAnimation ~= nil then
		if (CustomIdleAnimation.StopActiveIdleSetTemporarily) then
			CustomIdleAnimation.StopActiveIdleSetTemporarily()
		end
	end
end

local function InitializeAvailableEmotes()
	local cVarString = GetCVar("language.2")
	local ignoreCounterList = {}

	local function IsEmoteIgnored(emoteDataEntry, slashName)
		if not emoteDataEntry.Ignore then return false end

		local ignore = emoteDataEntry.Ignore
		if #ignore == 0 then return true end

		local count = ignoreCounterList[slashName]
		if not count then
			count = 1
			ignoreCounterList[slashName] = 1
		else
			count = count + 1
			ignoreCounterList[slashName] = count
		end
		return ignore[count] == true
	end

	local function AddEmoteToLists(emote)
		table.insert(_availableEmotes, emote)
		_idEmoteList[emote.ID] = emote
	end

	local function TryAddEmote(index)
		local slashName, categoryId, id, displayName = GetEmoteInfo(index)
		local emoteDataEntry = _emoteData[slashName]

		if not emoteDataEntry or IsEmoteIgnored(emoteDataEntry, slashName) == false then
			if emoteDataEntry ~= nil and emoteDataEntry.ReplaceName ~= nil then
				local newDisplayName = emoteDataEntry.ReplaceName[cVarString]
				if newDisplayName ~= nil then displayName = newDisplayName end
			end

			local emote = {
				SlashName = slashName,
				DisplayName = ZO_CachedStrFormat(SI_COLLECTIBLE_NAME_FORMATTER, displayName),
				CategoryID = categoryId,
				ID = id,
				Index = index,
				Play = function()
					PlayEmoteByIndex(index)
					ConcurrencyWorkaround()
				end,
				TagString = string.lower(zo_strformat("<<1>> <<2>>", slashName, displayName)),
			}

			AddEmoteToLists(emote)

			if not _categoryNames[categoryId] then
				_categoryNames[categoryId] = GetString(_G[zo_strformat("SI_EMOTECATEGORY<<1>>", categoryId)])
			end
		end
	end

	local numEmotes = GetNumEmotes()
	if not _savedAccountVariables.ShowLockedEmotes then
		for i = 1, numEmotes do
			local collectibleId = GetEmoteCollectibleId(i)
			if not collectibleId or IsCollectibleUnlocked(collectibleId) then
				TryAddEmote(i)
			end
		end
	else
		for i = 1, numEmotes do
			TryAddEmote(i)
		end
	end

	_numberOfAvailableEmotes = #_availableEmotes

	if _savedAccountVariables.EnablePersonalitiesAsFavorites then
		for i, data in ipairs(_availablePersonalities) do
			local emote = {
				SlashName = data.Name,
				DisplayName = data.Name,
				CategoryID = LE_Const_PersonalityCategoryId,
				ID = LE_Const_PersonalityOffset - data.CollectibleId,
				Play = function() UseCollectible(data.CollectibleId) end,
				TagString = string.lower(data.Name),
			}

			AddEmoteToLists(emote)
		end

		_categoryNames[LE_Const_PersonalityCategoryId] = GetString(SI_COLLECTIBLECATEGORYTYPE9)
	end

	local favoriteCommandsData = _savedAccountVariables.FavoriteCommandsData
	if #favoriteCommandsData > 0 then
		for id, data in ipairs(favoriteCommandsData) do
			local commandString = data.Command
			local displayName = data.DisplayName

			if displayName == nil then
				displayName = commandString
			end

			local emote = {
				SlashName = commandString,
				DisplayName = displayName,
				CategoryID = LE_Const_FavoriteCommandsCategoryId,
				ID = id * -1,
				Play = function() DoCommand(commandString) end,
				TagString = string.lower(zo_strformat("<<1>> <<2>>", commandString, displayName)),
			}

			AddEmoteToLists(emote)
		end

		_categoryNames[LE_Const_FavoriteCommandsCategoryId] = GetString(SI_GRAPHICSPRESETS7)
	end

	SortAvailableEmotes()
	RefreshAvailableEmotesOverridden()
end

local function InitializeAvailablePersonalities()
	local index = 1

	while true do
		local collectibleId = GetCollectibleIdFromType(COLLECTIBLE_CATEGORY_TYPE_PERSONALITY, index)
		if not collectibleId or collectibleId < 1 then break end

		if IsCollectibleUnlocked(collectibleId) then
			local personalityData = {
				CollectibleId = collectibleId,
				Name = ZO_CachedStrFormat(SI_COLLECTIBLE_NAME_FORMATTER, GetCollectibleName(collectibleId)),
			}

			table.insert(_availablePersonalities, personalityData)
		end

		index = index + 1
	end
end

local function GetActiveFavoriteEmotesTabIndex()
	if _savedLocalVariables.IsUsingSharedFavorites then
		return _savedAccountVariables.FavoriteEmotesTabIndex
	end

	return _savedLocalVariables.FavoriteEmotesTabIndex
end
LovelyEmotes.GetActiveFavoriteEmotesTabIndex = GetActiveFavoriteEmotesTabIndex

local function SetActiveFavoriteEmotesTabIndex(index)
	if _savedLocalVariables.IsUsingSharedFavorites then
		_savedAccountVariables.FavoriteEmotesTabIndex = index
	else
		_savedLocalVariables.FavoriteEmotesTabIndex = index
	end
end
LovelyEmotes.SetActiveFavoriteEmotesTabIndex = SetActiveFavoriteEmotesTabIndex

local function GetSavedFavoriteTab(index)
	if _savedLocalVariables.IsUsingSharedFavorites then
		return _savedAccountVariables.FavoriteEmotesTabs[index]
	end

	return _savedLocalVariables.FavoriteEmotesTabs[index]
end
LovelyEmotes.GetSavedFavoriteTab = GetSavedFavoriteTab

function LovelyEmotes.GetNumberOfVisibleFavoriteEmotesTabs()
	if _savedLocalVariables.IsUsingSharedFavorites then
		return _savedAccountVariables.NumberOfFavoriteEmotesTabs
	end

	return _savedLocalVariables.NumberOfFavoriteEmotesTabs
end

function LovelyEmotes.SetNumberOfVisibleFavoriteEmotesTabs(value)
	if _savedLocalVariables.IsUsingSharedFavorites then
		_savedAccountVariables.NumberOfFavoriteEmotesTabs = value
	else
		_savedLocalVariables.NumberOfFavoriteEmotesTabs = value
	end
end

function LovelyEmotes.GetFavoriteButtonCount(tabIndex)
	if tabIndex == nil then tabIndex = GetActiveFavoriteEmotesTabIndex() end
	return GetSavedFavoriteTab(tabIndex).ButtonCount
end

function LovelyEmotes.SetFavoriteButtonCount(value, tabIndex)
	if tabIndex == nil then tabIndex = GetActiveFavoriteEmotesTabIndex() end
	GetSavedFavoriteTab(tabIndex).ButtonCount = value
end

local function GetSavedID(emoteIndex, tabIndex)
	if tabIndex == nil then tabIndex = GetActiveFavoriteEmotesTabIndex() end
	return GetSavedFavoriteTab(tabIndex).EmoteIDs[emoteIndex]
end
LovelyEmotes.GetSavedID = GetSavedID

local function GetEmoteByID(id)
	if not id then return id end
	return _idEmoteList[id]
end
LovelyEmotes.GetEmoteByID = GetEmoteByID

local function GetSavedEmote(emoteIndex, tabIndex)
	return GetEmoteByID(GetSavedID(emoteIndex, tabIndex))
end
LovelyEmotes.GetSavedEmote = GetSavedEmote

function LovelyEmotes.GetCategoryNameByID(categoryId)
	return _categoryNames[categoryId]
end

local function SaveEmote(slotIndex, emoteID, tabIndex)
	if tabIndex == nil then tabIndex = GetActiveFavoriteEmotesTabIndex() end
	GetSavedFavoriteTab(tabIndex).EmoteIDs[slotIndex] = emoteID
end
LovelyEmotes.SaveEmote = SaveEmote

function LovelyEmotes.SaveRandomEmote(slotIndex, tabIndex)
	local randomEmoteIndex = math.random(1, #LovelyEmotes.AvailableEmotes)
	local emote = LovelyEmotes.AvailableEmotes[randomEmoteIndex]

	SaveEmote(slotIndex, emote.ID, tabIndex)

	return emote
end

function LovelyEmotes.GetRadialMenuTargetTabIndex()
	if _savedLocalVariables.IsUsingSharedFavorites then
		return _savedAccountVariables.RadialMenuTargetTabIndex
	end

	return _savedLocalVariables.RadialMenuTargetTabIndex
end

function LovelyEmotes.SetRadialMenuTargetTabIndex(value)
	if _savedLocalVariables.IsUsingSharedFavorites then
		_savedAccountVariables.RadialMenuTargetTabIndex = value
	else
		_savedLocalVariables.RadialMenuTargetTabIndex = value
	end
end

function LovelyEmotes.IsEmoteSynchronizationActive()
	return _isEmoteSynchronizationActive
end

function LovelyEmotes.PlaySavedEmote(index)
	local emote = GetSavedEmote(index)

	if not emote then
		ZO_Alert(UI_ALERT_CATEGORY_ALERT, nil, GetString(SI_LOVELYEMOTES_ALERT_NO_EMOTE_SELECTED))
		return
	end

	emote.Play()
end

function LovelyEmotes.SetAccountWideFavoritesActive(isActive)
	if _savedLocalVariables.IsUsingSharedFavorites == isActive then return end
	_savedLocalVariables.IsUsingSharedFavorites = isActive

	LovelyEmotes.MainWindow.ShowFavoriteEmotesTab(GetActiveFavoriteEmotesTabIndex())
end

local function AddLoopIcon(key, name)
	local data = _emoteData[key]
	if data == nil or not data.HasLoop then return name end
	return zo_strformat("<<1>> ∞", name)
end

function LovelyEmotes.GetEmoteDisplayName(emote)
	if not _savedAccountVariables.ShowEndlessLoopIcons then return emote.DisplayName end
	return AddLoopIcon(emote.SlashName, emote.DisplayName)
end

function LovelyEmotes.GetEmoteSlashName(emote)
	if not _savedAccountVariables.ShowEndlessLoopIcons then return emote.SlashName end
	return AddLoopIcon(emote.SlashName, emote.SlashName)
end

function LovelyEmotes.GetEmoteTextColor(emote)
	if emote then
		if emote.ID > -1 and emote.IsOverridden then
			return ZO_PERSONALITY_EMOTES_COLOR:UnpackRGBA()
		elseif emote.ID < LE_Const_PersonalityOffset and IsCollectibleActive(emote.ID * -1 + LE_Const_PersonalityOffset) then
			return LOVELYEMOTES_PERSONALITY_ACTIVE_COLOR:UnpackRGBA()
		end
	end

	return ZO_NORMAL_TEXT:UnpackRGBA()
end

function LovelyEmotes.ReinitializeAvailableEmotes()
	ZO_ClearNumericallyIndexedTable(_availableEmotes)
	ZO_ClearNumericallyIndexedTable(_availablePersonalities)
	ZO_ClearTable(_idEmoteList)

	InitializeAvailablePersonalities()
	InitializeAvailableEmotes()

	_mainWindow.RefreshAllFavoriteButtons()
	LovelyEmotes.EmoteList:ResetList()
end

local function SetWindowPositionsLocked(value)
	if LE_MainWindowControl:IsMouseEnabled() ~= value then return end
	_savedAccountVariables.LockedWindowPositions = value

	LE_MainWindowControl:SetMouseEnabled(not value)
	LE_MinimizedWindowControl:SetMouseEnabled(not value)
	LE_EmoteSelectionWindow:SetMouseEnabled(not value)

	local button = LE_MainWindowControlLockButton

	if value then
		button:SetNormalTexture("esoui/art/miscellaneous/locked_up.dds")
		button:SetMouseOverTexture("esoui/art/miscellaneous/locked_over.dds")
		button:SetPressedTexture("esoui/art/miscellaneous/locked_down.dds")
		button:SetDisabledTexture("esoui/art/miscellaneous/locked_up.dds")
		return
	end

	button:SetNormalTexture("esoui/art/miscellaneous/unlocked_up.dds")
	button:SetMouseOverTexture("esoui/art/miscellaneous/unlocked_over.dds")
	button:SetPressedTexture("esoui/art/miscellaneous/unlocked_down.dds")
	button:SetDisabledTexture("esoui/art/miscellaneous/unlocked_up.dds")
end
LovelyEmotes.SetWindowPositionsLocked = SetWindowPositionsLocked

function LovelyEmotes.SetCombatStateEventActive(value)
	_savedAccountVariables.MinimizeInCombat = value

	if value then
		EVENT_MANAGER:RegisterForEvent(_name, EVENT_PLAYER_COMBAT_STATE, OnCombatStateChanged)
		return
	end

	EVENT_MANAGER:UnregisterForEvent(_name, EVENT_PLAYER_COMBAT_STATE)
end

local function ModifyEmoteData()
	_emoteData = LovelyEmotes.EmoteData

	local emoteDataMod = LovelyEmotes.Modifications.EmoteDataMod
	if not emoteDataMod then return end

	local cVarString = GetCVar("language.2")

	for key, entry in pairs(emoteDataMod) do
		if not _emoteData[key] then _emoteData[key] = {} end

		if entry.Ignore ~= nil then
			_emoteData[key].Ignore = entry.Ignore
		end

		if entry.ReplaceName ~= nil and entry.ReplaceName[cVarString] ~= nil then
			if not _emoteData[key].ReplaceName then
				_emoteData[key].ReplaceName = {}
			end

			_emoteData[key].ReplaceName[cVarString] = entry.ReplaceName[cVarString]
		end
	end
end

local function UpdateEmotes()
	RefreshAvailableEmotesOverridden()
	_mainWindow.RefreshAllFavoriteButtons()
	LovelyEmotes.EmoteList:RefreshVisible()
end

local function LeOnEventPersonalityChanged()
	UpdateEmotes()
	_personalityList.RefreshVisible()
end

--[[
local function LeOnCollectibleUpdated(collectibleId)
	local collectibleData = ZO_COLLECTIBLE_DATA_MANAGER:GetCollectibleDataById(collectibleId)
	if not collectibleData:IsCategoryType(COLLECTIBLE_CATEGORY_TYPE_EMOTE) or collectibleData:IsCategoryType(COLLECTIBLE_CATEGORY_TYPE_PERSONALITY) then return end

	UpdateEmotes()
	_personalityList.RefreshVisible()
end
]]

local function LeOnCollectionUpdated(collectionUpdateType, collectiblesByNewUnlockState)
	for k, unlockStates in pairs(collectiblesByNewUnlockState) do
		for i, collectibleData in ipairs(unlockStates) do
			if collectibleData:IsCategoryType(COLLECTIBLE_CATEGORY_TYPE_EMOTE) or collectibleData:IsCategoryType(COLLECTIBLE_CATEGORY_TYPE_PERSONALITY) then
				LovelyEmotes.ReinitializeAvailableEmotes()
				_personalityList.Reset()
				return
			end
		end
	end
end

local function OnChatMessage(eventCode, channelType, fromName, text, isCustomerService, fromDisplayName)
	if channelType == CHAT_CHANNEL_SAY then if not _savedAccountVariables.EnableSayChatSync then return end
	elseif channelType == CHAT_CHANNEL_PARTY then if not _savedAccountVariables.EnablePartyChatSync then return end
	elseif channelType == CHAT_CHANNEL_WHISPER or channelType == CHAT_CHANNEL_WHISPER_SENT then if not _savedAccountVariables.EnableWhisperChatSync then return end
	elseif channelType == CHAT_CHANNEL_ZONE then if not _savedAccountVariables.EnableZoneChatSync then return end
	else return end

	local strings = {}
	for w in string.gmatch(text, "%S+") do
		table.insert(strings, w)
		if #strings > 1 then break end
	end

	if strings[1] ~= "LovelyEmotes" or strings[2] == nil then return end

	local emote = _idEmoteList[tonumber(strings[2]:match("%d+"))]
	if emote ~= nil then
		emote.Play()
	end
end

local function UpdateChatMessageChannelEventActive()
	local isActive = _savedAccountVariables.EnableSayChatSync or _savedAccountVariables.EnablePartyChatSync or _savedAccountVariables.EnableWhisperChatSync or _savedAccountVariables.EnableZoneChatSync
	if isActive == _isEmoteSynchronizationActive then return end
	_isEmoteSynchronizationActive = isActive

	if isActive then
		EVENT_MANAGER:RegisterForEvent(_name, EVENT_CHAT_MESSAGE_CHANNEL, OnChatMessage)
		return
	end

	EVENT_MANAGER:UnregisterForEvent(_name, EVENT_CHAT_MESSAGE_CHANNEL)
end
LovelyEmotes.UpdateChatMessageChannelEventActive = UpdateChatMessageChannelEventActive

local function OnAddonLoaded(event, addonName)
	if addonName ~= _name then return end
	EVENT_MANAGER:UnregisterForEvent(addonName, EVENT_ADD_ON_LOADED)

	_savedAccountVariables = ZO_SavedVars:NewAccountWide("LovelyEmotesVariables", 1, nil, LovelyEmotes.AccountVariables)
	_savedLocalVariables = ZO_SavedVars:NewCharacterIdSettings("LovelyEmotesVariables", 2, nil, LovelyEmotes.LocalVariables)

	if #_savedAccountVariables.FavoriteEmotesTabs[1].EmoteIDs < 1 then
		_savedAccountVariables.FavoriteEmotesTabs[1].EmoteIDs = { [1] = 74, [2] = 33, [3] = 120, [4] = 11, [5] = 5, }
	end

	LovelyEmotes.SavedAccountVariables = _savedAccountVariables
	LovelyEmotes.SavedLocalVariables = _savedLocalVariables

	local invisibleFragment = ZO_SimpleSceneFragment:New(LE_Invisible)
	HUD_SCENE:AddFragment(invisibleFragment)
	HUD_UI_SCENE:AddFragment(invisibleFragment)

	LovelyEmotes.UpdateSettings()
	LE_FavoriteCommands_RemoveEmptyData()
	ModifyEmoteData()
	InitializeAvailablePersonalities()
	InitializeAvailableEmotes()
	LovelyEmotes_EmoteList_ShowSlashNames(_savedAccountVariables.ShowSlashNames)

	LovelyEmotes.EmoteList = LovelyEmotes_EmoteList_CreateNew()
	_emoteSelectionWindow.Initialize()
	_mainWindow.Initialize()
	_personalityList.Initialize()
	LovelyEmotes.RadialMenu = LovelyEmotes_CreateRadialMenu()
	LovelyEmotes.FavoriteCommandsMenu = LovelyEmotes_CreateFavoriteCommandsMenu()
	LE_ChatWindowButton:Initialize()
	LovelyEmotes.InitializeSettingsMenu()

	LovelyEmotes.UpdateButtonClickSounds(_savedAccountVariables.EnableAlternativeSounds)

	SetWindowPositionsLocked(_savedAccountVariables.LockedWindowPositions)

	EVENT_MANAGER:RegisterForEvent(_name, EVENT_PERSONALITY_CHANGED, LeOnEventPersonalityChanged)
	--ZO_COLLECTIBLE_DATA_MANAGER:RegisterCallback("OnCollectibleUpdated", LeOnCollectibleUpdated)
	ZO_COLLECTIBLE_DATA_MANAGER:RegisterCallback("OnCollectionUpdated", LeOnCollectionUpdated)
	UpdateChatMessageChannelEventActive()
end

local function OnPlayerActivated(eventCode, initial)
	EVENT_MANAGER:UnregisterForEvent(_name, EVENT_PLAYER_ACTIVATED)

	EVENT_MANAGER:RegisterForEvent(_name, EVENT_GAME_CAMERA_UI_MODE_CHANGED, OnGameCameraUiModeChanged)
	if _savedAccountVariables.MinimizeInCombat then LovelyEmotes.SetCombatStateEventActive(true) end
end

local function OnRandomSlashCommand(arg)
	local targetEmotes = {}

	for i, emote in ipairs(_availableEmotes) do
		if emote.ID >= 0 then
			local collectibleId = GetEmoteCollectibleId(emote.Index)

			if not collectibleId then
				table.insert(targetEmotes, emote)
			elseif IsCollectibleUnlocked(collectibleId) then
				table.insert(targetEmotes, emote)
			end
		end
	end

	targetEmotes[math.random(1, #targetEmotes)].Play()
end

SLASH_COMMANDS["/letoggle"] = function(arg) _mainWindow.ToggleMinimized() end
SLASH_COMMANDS["/lerandom"] = OnRandomSlashCommand
SLASH_COMMANDS["/lefavoritecommands"] = function(arg) LovelyEmotes.FavoriteCommandsMenu:Open() end
EVENT_MANAGER:RegisterForEvent(_name, EVENT_ADD_ON_LOADED, OnAddonLoaded)
EVENT_MANAGER:RegisterForEvent(_name, EVENT_PLAYER_ACTIVATED, OnPlayerActivated)

--LOVELYEMOTES_CALLBACK_MANAGER = ZO_CallbackObject:Subclass():New()
